import numpy as np
from npqtools.QUBO.qubo import QUBO
from npqtools.problems.PrintGraph import PaintVertexSetInWeightedGraph

class QUBOMaxWeightClique(QUBO) :
    def __init__(self, adjastency_matrix : np.array):
        super().__init__()
        self.adjastency_matrix = adjastency_matrix
        self.adjastency_matrix_size = adjastency_matrix.shape[0]
        self.compute_qubo()

    def change_adjastency_matrix(self) :
        for i in range(self.adjastency_matrix_size) :
            self.adjastency_matrix[i][i] = 0

    def generate_mask(self) -> np.array :
        return np.array([[1 if self.adjastency_matrix[i][j] else 0 for j in range(self.adjastency_matrix_size)] for i in range(self.adjastency_matrix_size)])

    def compute_qubo(self):
        self.shape = self.adjastency_matrix_size
        self.change_adjastency_matrix()
        self.matrix = - self.adjastency_matrix
        mask = self.generate_mask()
        self.matrix += self.adjastency_matrix.sum() * (np.ones(self.matrix.shape) - mask - np.eye(self.matrix.shape[0])).astype(int)
    

    def set_solution(self) -> None:
        self.solution = np.array([key for key, value in self.raw_solution.items() if value])

    
    def display_solution(self) -> None :
        PaintVertexSetInWeightedGraph(self.adjastency_matrix, self.solution)



import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

def PaintNonOrientatedGraphWithWay (matrix : np.array, way : np.array) -> None :
    G = nx.Graph()
    G.add_nodes_from(list(range(1, matrix.shape[0] + 1)))
    for i in range(matrix.shape[0]) :
        for j in range(i, matrix.shape[0]) :
            if matrix[i][j] :
                G.add_weighted_edges_from([(i + 1, j + 1, matrix[i][j])])
    Positions = nx.spring_layout(G)
    nx.draw_networkx_nodes(G, Positions, node_color="black", node_size=100)
    Weights = nx.get_edge_attributes(G, "weight")
    nx.draw_networkx_edges(G, Positions, width=2, alpha=0.5, edge_color="blue")
    EdgeLabels = {edge : f"{Weights[edge]}" for edge in G.edges()}
    HighlightedEdges = []
    for i in range(1, way.size) :
        HighlightedEdges.append((way[i - 1] + 1, way[i] + 1))
    nx.draw_networkx_edges(G, Positions, edgelist=HighlightedEdges, edge_color="green", width=3)
    nx.draw_networkx_edge_labels(G, Positions, edge_labels=EdgeLabels, font_color="red")
    plt.title("Our Task Graph Visualisation")
    plt.axis("off")
    plt.show()
    return


def PaintNonOrientatedGraphWithCycle (matrix : np.array, way : np.array) -> None :
    G = nx.Graph()
    G.add_nodes_from(list(range(1, matrix.shape[0] + 1)))
    for i in range(matrix.shape[0]) :
        for j in range(i, matrix.shape[0]) :
            if matrix[i][j] :
                G.add_weighted_edges_from([(i + 1, j + 1, matrix[i][j])])
    Positions = nx.spring_layout(G)
    nx.draw_networkx_nodes(G, Positions, node_color="black", node_size=100)
    Weights = nx.get_edge_attributes(G, "weight")
    nx.draw_networkx_edges(G, Positions, width=2, alpha=0.5, edge_color="blue")
    EdgeLabels = {edge : f"{Weights[edge]}" for edge in G.edges()}
    HighlightedEdges = []
    for i in range(1, way.size + 1) :
        HighlightedEdges.append((way[i - 1] + 1, way[i % way.size] + 1))
    nx.draw_networkx_edges(G, Positions, edgelist=HighlightedEdges, edge_color="green", width=3)
    nx.draw_networkx_edge_labels(G, Positions, edge_labels=EdgeLabels, font_color="red")
    plt.title("Our Task Graph Visualisation")
    plt.axis("off")
    plt.show()
    return


def PaintVertexSetInNonWeightedGraph (matrix : np.array, vertex_list : np.array) -> None :
    G = nx.Graph()
    G.add_nodes_from(list(range(1, matrix.shape[0] + 1)))
    for i in range(matrix.shape[0]) :
        for j in range(i + 1, matrix.shape[0]) :
            if matrix[i][j] :
                G.add_weighted_edges_from([(i + 1, j + 1, matrix[i][j])])
    Positions = nx.spring_layout(G)
    for i in range(vertex_list.size) :
        vertex_list[i] += 1
    nx.draw_networkx_nodes(G, Positions, node_color="black", node_size=100)
    nx.draw_networkx_nodes(G, Positions, node_color="red", node_size=100, nodelist=vertex_list)
    nx.draw_networkx_edges(G, Positions, width=2, alpha=0.5, edge_color="blue")
    HighlightedEdges = []
    for i in range(vertex_list.size):
        for j in range(i + 1, vertex_list.size) :
            HighlightedEdges.append((vertex_list[i], vertex_list[j]))
    nx.draw_networkx_edges(G, Positions, edgelist=HighlightedEdges, edge_color="red", width=3)
    plt.title("Our Clique Visualisation")
    plt.axis("off")
    plt.show()
    return



def PaintVertexSetInWeightedGraph (matrix : np.array, vertex_list : np.array) -> None :
    G = nx.Graph()
    G.add_nodes_from(list(range(1, matrix.shape[0] + 1)))
    for i in range(matrix.shape[0]) :
        for j in range(i, matrix.shape[0]) :
            if matrix[i][j] :
                G.add_weighted_edges_from([(i + 1, j + 1, matrix[i][j])])
    Positions = nx.spring_layout(G)
    for i in range(vertex_list.size) :
        vertex_list[i] += 1
    nx.draw_networkx_nodes(G, Positions, node_color="black", node_size=100)
    nx.draw_networkx_nodes(G, Positions, node_color="red", node_size=150, nodelist=vertex_list)
    nx.draw_networkx_edges(G, Positions, width=2, alpha=0.5, edge_color="blue")
    HighlightedEdges = []
    for i in range(vertex_list.size):
        for j in range(i + 1, vertex_list.size) :
            HighlightedEdges.append((vertex_list[i], vertex_list[j]))
    nx.draw_networkx_edges(G, Positions, edgelist=HighlightedEdges, edge_color="red", width=3)
    Weights = nx.get_edge_attributes(G, "weight")
    EdgeLabels = {edge : f"{Weights[edge]}" for edge in G.edges()}
    nx.draw_networkx_edge_labels(G, Positions, edge_labels=EdgeLabels, font_color="black", font_size=7)
    plt.title("Our Clique Visualisation")
    plt.axis("off")
    plt.show()
    return


from npqtools.QUBO.qubo import QUBO
import numpy as np


class QUBOKnapsack(QUBO):
    def __init__(self, objects_list, capability):
        super().__init__()
        self.objects_weights = objects_list[:, 1].reshape(-1, 1)
        self.objects_costs = objects_list[:, 0]
        self.capability = capability
        self.sum_weight = 0
        self.compute_qubo()

    def set_solution(self):
        self.solution = np.array([key for key, value in self.raw_solution.items() if value == np.int8(1)])
        self.solution = self.solution[self.solution < self.objects_weights.shape[0]]
        self.sum_weight = sum(map(lambda obj_numb: self.objects_weights[obj_numb], self.solution))[0]

    def _int_log(self, n: int) -> int: #M : 2^M <= n < 2^(M+1)
        res_log = 0
        value = 1
        while value <= n:
            res_log += 1
            value *= 2
        return res_log - 1

    def compute_qubo(self):
        A = self.objects_costs.max()
        M = self._int_log(self.capability)
        self.shape = self.objects_weights.shape[0] + 1 + M

        # Разобьем первое слагаемое на 4 матрицы

        degs = np.hstack([2 ** np.arange(M), self.capability + 1 - 2 ** M]).reshape(-1, 1)
        matrix1 = self.objects_weights @ self.objects_weights.T
        matrix2 = -(self.objects_weights @ degs.T)
        matrix3 = -(degs @ self.objects_weights.T)
        matrix4 = degs @ degs.T

        self.matrix = A * np.vstack([np.hstack([matrix1, matrix2]), np.hstack([matrix3, matrix4])]) - np.diag(np.hstack([self.objects_costs, np.zeros(M + 1)])).astype(int)
        self.delta = 0

    def display_solution(self):
        print(f"Наилучшая сумма: {-self.min_energy}")
        print(f"Суммарный вес: {self.sum_weight}")
        print("Товары с номерами:")
        print(self.solution)


from npqtools.problems.QuboClique import QUBOClique
from npqtools.problems.QuboHam import QUBOHam
from npqtools.problems.QuboSalesman import QUBOSalesman
from npqtools.problems.QuboKnapsack import QUBOKnapsack
from npqtools.problems.QuboMaxWeightClique import QUBOMaxWeightClique

__all__ = ["QUBOClique", "QUBOHam", "QUBOSalesman", "QUBOKnapsack", "QUBOMaxWeightClique"]


import numpy as np
from npqtools.QUBO.qubo import QUBO
from npqtools.problems.PrintGraph import PaintVertexSetInNonWeightedGraph

class QUBOClique(QUBO) :
    def __init__ (self, adjastency_matrix : np.array) :
        super().__init__()
        self.adjastency_matrix = adjastency_matrix
        self.adjastency_matrix_size = adjastency_matrix.shape[0]
        self.compute_qubo()

    def change_adjstency_matrix(self) -> None :
        for i in range(self.adjastency_matrix_size) :
            for j in range(self.adjastency_matrix_size) :
                if self.adjastency_matrix[i][j] :
                    self.adjastency_matrix[i][j] = 1
                if i == j :
                    self.adjastency_matrix[i][j] = 0

    def compute_qubo(self) -> None :
        self.change_adjstency_matrix()
        self.shape = self.adjastency_matrix_size
        self.matrix = np.zeros(self.adjastency_matrix.shape)
        self.matrix -= np.eye(self.adjastency_matrix.shape[0])
        self.matrix += self.adjastency_matrix.sum() * (np.ones(self.adjastency_matrix.shape) - self.adjastency_matrix - np.eye(self.adjastency_matrix.shape[0]))

    def set_solution(self) -> None:
        self.solution = np.array([key for key, value in self.raw_solution.items() if value])

    def display_solution(self) -> None :
        PaintVertexSetInNonWeightedGraph(self.adjastency_matrix, self.solution)




from npqtools.QUBO.qubo import QUBO
from npqtools.problems.PrintGraph import PaintNonOrientatedGraphWithCycle
import numpy as np


class QUBOHam(QUBO):
    def __init__(self, adjastency_matrix):
        super().__init__()
        self.adjastency_matrix = adjastency_matrix
        self.adjastency_matrix_size = adjastency_matrix.shape[0]
        self.compute_qubo()

    #Возвращает гамильтонов путь
    def set_solution(self):
        solution = np.array([key % self.adjastency_matrix_size for key, value in self.raw_solution.items() if value == np.int8(1)])
        self.solution = solution if solution.shape[0] == self.adjastency_matrix_size else None

    def compute_qubo(self):
        self.shape = self.adjastency_matrix_size ** 2

        indices = np.arange(self.shape).reshape(-1, 1)

        mask1 = ((indices // self.adjastency_matrix_size) == (indices.T // self.adjastency_matrix_size)).astype(int) - 2 * np.eye(
            self.shape)
        mask2 = ((indices % self.adjastency_matrix_size) == (indices.T % self.adjastency_matrix_size)).astype(int) - 2 * np.eye(
            self.shape)
        mask3 = ((self.adjastency_matrix[indices // self.adjastency_matrix_size, indices.T // self.adjastency_matrix_size] == 0) & ((
            indices % self.adjastency_matrix_size - indices.T % self.adjastency_matrix_size + 1) % self.adjastency_matrix_size == 0)).astype(
            int)

        self.matrix = mask1 + mask2 + mask3
        self.delta = 2 * self.adjastency_matrix_size

    def display_solution(self):
        PaintNonOrientatedGraphWithCycle(self.adjastency_matrix, self.solution) if self.solution is not None else print("Nothing to show!")
